@@ -448,6 +448,7 @@ bool buildHash(const std::string& outputDir, nonstd::string_view concatText,
 template <typename ParserT> //, typename CoverageCalculator>
 void indexTranscriptsSA(ParserT* parser,
                         std::string& outputDir,
+                        spp::sparse_hash_set<std::string>& decoyNames,
                         bool noClipPolyA, bool usePerfectHash,
                         uint32_t numHashThreads,
                         bool keepDuplicates,
@@ -467,10 +468,21 @@ void indexTranscriptsSA(ParserT* parser,
   //picosha2::hash256_one_by_one seqHasher; seqHasher.init();
   //picosha2::hash256_one_by_one nameHasher; nameHasher.init();
 
+  bool firstRecord{true};
+  bool hasGencodeSep = (sepStr.find('|') != std::string::npos);
   uint32_t n{0};
   uint32_t k = rapmap::utils::my_mer::k();
   std::vector<std::string> transcriptNames;
   std::vector<int64_t> transcriptStarts;
+
+  // Keep track of if we've seen a decoy sequence yet.
+  // The index enforces that all decoy sequences are consecutive, and that
+  // they come after all valid (non-decoy) sequences.  If we see a non-decoy
+  // sequence after having observed a decoy, then we complain and exit.
+  bool sawDecoy{false};
+  uint64_t numberOfDecoys{0};
+  uint64_t firstDecoyIndex{std::numeric_limits<uint64_t>::max()};
+
   // std::vector<uint32_t> positionIDs;
   constexpr char bases[] = {'A', 'C', 'G', 'T'};
   uint32_t polyAClipLength{10};
@@ -482,6 +494,7 @@ void indexTranscriptsSA(ParserT* parser,
   using KmerBinT = uint64_t;
 
   bool clipPolyA = !noClipPolyA;
+  bool haveDecoys = !decoyNames.empty();
 
   struct DupInfo {
     uint64_t txId;
@@ -531,6 +544,34 @@ void indexTranscriptsSA(ParserT* parser,
 
         // get the hash to check for collisions before we change anything.
         auto txStringHash = XXH64(reinterpret_cast<void*>(const_cast<char*>(readStr.data())), readLen, 0);
+        auto& readName = read.name;
+
+        // check if we think this is a gencode transcriptome, and the user has not passed the gencode flag
+        if (firstRecord and !hasGencodeSep) {
+          constexpr const size_t numGencodeSep{8};
+          if ( std::count(readName.begin(), readName.end(), '|') == numGencodeSep ) {
+            log->warn("It appears that this may be a GENCODE transcriptome (from analyzing the separators in the FASTA header).  However, "
+                      "you have not set \'|\' as a header separator.  If this is a GENCODE transcriptome, consider passing --gencode to the "
+                      "salmon index command.\n\n");
+          }
+          firstRecord = false;
+        }
+
+        bool isDecoy = (haveDecoys) ? decoyNames.contains(readName) : false;
+        // If this is *not* a decoy sequence, make sure that
+        // we haven't seen any decoys yet.  Otherwise we are violating
+        // the condition that decoys must come last.
+        if (!isDecoy and sawDecoy) {
+          log->critical("Observed a non-decoy sequence [{}] after having already observed a decoy. "
+                        "However, it is required that any decoy target records appear, consecutively, "
+                        "at the end of the input fasta file.  Please re-format your input file so that "
+                        "all decoy records appear contiguously at the end of the file, after all valid "
+                        "(non-decoy) records", readName);
+          log->flush();
+          spdlog::drop_all();
+          std::exit(1);
+          //throw std::logic_error("Input fasta file contained out-of-order decoy targets.");
+        }
 
         // First, replace non ATCG nucleotides
         for (size_t b = 0; b < readLen; ++b) {
@@ -570,7 +611,7 @@ void indexTranscriptsSA(ParserT* parser,
         if (readStr.size() > 0) {
           // If we're suspicious the user has fed in a *genome* rather
           // than a transcriptome, say so here.
-          if (readStr.size() >= tooLong) {
+          if (readStr.size() >= tooLong and !isDecoy) {
             log->warn("Entry with header [{}] was longer than {} nucleotides.  "
                       "Are you certain that "
                       "we are indexing a transcriptome and not a genome?",
@@ -622,6 +663,18 @@ void indexTranscriptsSA(ParserT* parser,
           transcriptStarts.push_back(currIndex);
           // The un-molested length of this transcript
           completeLengths.push_back(completeLen);
+          if (isDecoy) {
+            // if we haven't seen another decoy yet, this is the first decoy
+            // index
+            if (!sawDecoy) {
+              firstDecoyIndex = txpIndex;
+            }
+            // once we see the first decoy, saw decoy is set to true
+            // for the rest of the processing.
+            sawDecoy = true;
+            ++numberOfDecoys;
+            //decoyIndices.push_back(txpIndex);
+          }
 
           // If we made it here, we were not an actual duplicate, so add this transcript
           // for future duplicate checking.
@@ -708,6 +761,9 @@ void indexTranscriptsSA(ParserT* parser,
     log->info("Writing sequence data to file . . . ");
     cereal::BinaryOutputArchive seqArchive(seqStream);
     seqArchive(transcriptNames);
+    seqArchive(numberOfDecoys);
+    seqArchive(firstDecoyIndex);
+
     if (largeIndex) {
       seqArchive(transcriptStarts);
     } else {
@@ -731,6 +787,21 @@ void indexTranscriptsSA(ParserT* parser,
   }
   seqStream.close();
 
+  /*
+  // Save the bit vector that tells us what targets are decoys
+  BIT_ARRAY* decoyArray = bit_array_create(transcriptNames.size());
+  if (haveDecoys) {
+    for (auto i : decoyIndices) {
+      bit_array_set_bit(decoyArray, i);
+    }
+  }
+  std::string decoyBVFileName = outputDir + "isdecoy.bin";
+  FILE* decoyBVFile = fopen(decoyBVFileName.c_str(), "w");
+  bit_array_save(decoyArray, decoyBVFile);
+  fclose(decoyBVFile);
+  bit_array_free(decoyArray);
+  */
+
   // clear stuff we no longer need
   // positionIDs.clear();
   // positionIDs.shrink_to_fit();
@@ -788,7 +859,7 @@ void indexTranscriptsSA(ParserT* parser,
   //seqHasher.finish();
   //nameHasher.finish();
 
-  std::string indexVersion = "q5";
+  std::string indexVersion = "q6";
   IndexHeader header(IndexType::QUASI,
                      indexVersion,
                      true, k, largeIndex,
@@ -818,6 +889,22 @@ void indexTranscriptsSA(ParserT* parser,
   headerStream.close();
 }
 
+spp::sparse_hash_set<std::string> populateDecoyHash(const std::string& fname, std::shared_ptr<spdlog::logger> log) {
+  spp::sparse_hash_set<std::string> dset;
+  std::ifstream dfile(fname);
+
+  std::string dname;
+  while (dfile >> dname) {
+    auto it = dset.insert(dname);
+    if (!it.second) {
+      log->warn("The decoy name {} was encountered more than once --- please be sure all decoy names and sequences are unique.", dname);
+    }
+  }
+
+  dfile.close();
+  return dset;
+}
+
 int rapMapSAIndex(int argc, char* argv[]) {
   TCLAP::CmdLine cmd("RapMap Indexer");
   TCLAP::ValueArg<std::string> transcripts("t", "transcripts",
@@ -836,6 +923,9 @@ int rapMapSAIndex(int argc, char* argv[]) {
   TCLAP::SwitchArg noClip(
       "n", "noClip",
       "Don't clip poly-A tails from the ends of target sequences", false);
+  TCLAP::ValueArg<std::string> decoyList(
+                             "d", "decoys",
+                             "Treat these sequences as decoys that may have sequence homologous to some known transcript", false, "", "path");
   TCLAP::SwitchArg perfectHash(
       "p", "perfectHash", "Use a perfect hash instead of sparse hash --- "
                           "somewhat slows construction, but uses less memory",
@@ -859,6 +949,7 @@ int rapMapSAIndex(int argc, char* argv[]) {
   cmd.add(perfectHash);
   cmd.add(customSeps);
   cmd.add(numHashThreads);
+  cmd.add(decoyList);
   cmd.parse(argc, argv);
 
   // stupid parsing for now
@@ -899,10 +990,19 @@ int rapMapSAIndex(int argc, char* argv[]) {
   std::vector<spdlog::sink_ptr> sinks{consoleSink, fileSink};
   auto jointLog = spdlog::create("jointLog", std::begin(sinks), std::end(sinks));
 
+  spp::sparse_hash_set<std::string> decoyNames;
+  if (decoyList.isSet()) {
+    std::string decoyFileName = decoyList.getValue();
+    bool decoyFileExists = rapmap::fs::FileExists(decoyFileName.c_str());
+    if (!decoyFileExists) {
+      jointLog->error("The decoy file {} does not exist.", decoyFileName);
+      std::exit(1);
+    }
+    decoyNames = populateDecoyHash(decoyFileName, jointLog);
+  }
+
   size_t numThreads{1};
-
   std::unique_ptr<single_parser> transcriptParserPtr{nullptr};
-
   size_t numProd = 1;
   transcriptParserPtr.reset(
 			    new single_parser(transcriptFiles, numThreads, numProd));
@@ -912,7 +1012,7 @@ int rapMapSAIndex(int argc, char* argv[]) {
   bool keepDuplicates = keepDuplicatesSwitch.getValue();
   uint32_t numPerfectHashThreads = numHashThreads.getValue();
   std::mutex iomutex;
-  indexTranscriptsSA(transcriptParserPtr.get(), indexDir, noClipPolyA,
+  indexTranscriptsSA(transcriptParserPtr.get(), indexDir, decoyNames, noClipPolyA,
                      usePerfectHash, numPerfectHashThreads, keepDuplicates, sepStr, iomutex, jointLog);
 
   // Output info about the reference
